from vhr_cloudmask.model.pipelines.cloudmask_cnn_pipeline \
    import CloudMaskPipeline

__author__ = """Jordan A. Caraballo-Vega"""
__email__ = "jordan.a.caraballo-vega@nasa.gov"
__version__ = "1.3.2"

__all__ = ["CloudMaskPipeline"]
